/* 
	 Author: ranmses@gmail.com
	 
	 Test case generator for the Travelling Salesman with
	 Multiple Passengers and High Occupancy Problem (TSMPHOP).
*/ 
#include <algorithm>
#include <fstream>
#include <iomanip>
#include <iosfwd>
#include <iostream>
#include <random>
#include <sstream>
#include <stdlib.h>
#include <unistd.h>
#include <vector>
// #include <bits/stdc++.h>
	 
using namespace std;

#define SHOW(variable) cout << #variable": "<< (variable) << std::endl;

typedef vector<int> vi;
typedef vector<vi> vvi;
typedef pair<int, int> pi;
typedef vector<pi> vpi;

int INF = 99999999;
vvi graph; 

int N;

void markhere(){ cout<< "HERE!\n" ; cin.ignore(); }

/***************************************************************************
					Kruskal             
***************************************************************************/
// http://www.geeksforgeeks.org/kruskals-minimum-spanning-tree-using-stl-in-c/ 

// Structure to represent a graph
struct Graph
{
    int V, E;
    vector< pair<int, pi> > eList;
 
    // Constructor
    Graph(int V, int E)
    {
        this->V = V;
        this->E = E;
    }
 
    // Utility function to add an edge
    void addEdge(int u, int v, int w)
    {
        eList.push_back({w, {u, v}});
    }
 
    // Function to find MST using Kruskal's
    // MST algorithm
    int kruskalMST();
};
 
// To represent Disjoint Sets
struct DisjointSets
{
    int *parent, *rnk;
    int n;
 
    // Constructor.
    DisjointSets(int n)
    {
        // Allocate memory
        this->n = n;
        parent = new int[n+1];
        rnk = new int[n+1];
 
        // Initially, all vertices are in
        // different sets and have rank 0.
        for (int i = 0; i <= n; i++)
        {
            rnk[i] = 0;
 
            //every element is parent of itself
            parent[i] = i;
        }
    }
 
    // Find the parent of a node 'u'
    // Path Compression
    int find(int u)
    {
        /* Make the parent of the nodes in the path
           from u--> parent[u] point to parent[u] */
        if (u != parent[u])
            parent[u] = find(parent[u]);
        return parent[u];
    }
 
    // Union by rank
    void merge(int x, int y)
    {
        x = find(x), y = find(y);
 
        /* Make tree with smaller height
           a subtree of the other tree  */
        if (rnk[x] > rnk[y])
            parent[y] = x;
        else // If rnk[x] <= rnk[y]
            parent[x] = y;
 
        if (rnk[x] == rnk[y])
            rnk[y]++;
    }
};
 
 /* Functions returns weight of the MST*/
 
int Graph::kruskalMST()
{
    int mst_wt = 0; // Initialize result
 
    // Sort eList in increasing order on basis of cost
    sort(eList.begin(), eList.end());
 
    // Create disjoint sets
    DisjointSets ds(V);
 
    // Iterate through all sorted eList
    vector< pair<int, pi> >::iterator it;
    // cout << "\nEdges of MST are \n";
    for (it=eList.begin(); it!=eList.end(); it++)
    {
        int u = it->second.first;
        int v = it->second.second;
 
        int set_u = ds.find(u);
        int set_v = ds.find(v);
 
        // Check if the selected edge is creating
        // a cycle or not (Cycle is created if u
        // and v belong to same set)
        if (set_u != set_v)
        {
            // Current edge will be in the MST
            // so print it
            // cout << u << " - " << v << endl;
 
            // Update MST weight
            mst_wt += it->first;
 
            // Merge two sets
            ds.merge(set_u, set_v);
        }
    }
 
    return mst_wt;
}

/***************************************************************************
					Auxiliary             
***************************************************************************/

int mersenne (int a, int b)
{
	//Mersernne-Twister
	random_device rd;
	mt19937 rng(rd());
	// uniform_int_distribution<int> uni(a,b);
	uniform_real_distribution<double> uni(a,b);
	// cout << uni(rng) << endl;
	return uni(rng);
}

void vviPrint (vvi &Array)
{
	// int Ordem = Array.size();
	for (int i = 0; i < Array.size(); ++i)
	{
		// cout << setw(2) << i << ":  | ";
		for (int j = 0; j < Array[i].size(); ++j)
		{
			//if (Array[i][j] > -0.000000001 && Array[i][j] <= 0) Array[i][j] = 0;
			//if (Array[i][j] == -0) Array[i][j] = 0;
			//Array[i][j] = roundf(Array[i][j] * 10000) / 10000;
			cout.precision(2);
			//cout << ((abs(Array[i][j]) < 0.000005)? 0.000: Array[i][j]) << endl;
			cout << setw(3) <<  Array[i][j] << " ";
		}
		// cout << "|" << endl;
		cout << endl;
	}
	cout << endl;
}

void cppPrint (int N, int C, int H, int P, vvi &e, vvi &t, vvi &p)
{
	cout << N << " " << C << " " << H << " " << P << endl << endl;
	vviPrint(e);
	vviPrint(t);
	vviPrint(p);
}

void mznvviPrint (vvi &Array)
{
	// int Ordem = Array.size();
	for (int i = 0; i < Array.size(); ++i)
	{
		if ( i == 0 )
			cout << "[| ";
		else
			cout << setw(7) << " ";
		for (int j = 0; j < Array[i].size(); ++j)
		{
			cout.precision(2);
			cout << setw(3) << Array[i][j];
			if ( j != Array[i].size()-1 )
				cout << ", ";
			else
				cout << " |";
		}
		if ( i != Array.size()-1 )
			cout << endl;
		else
			cout << "];" << endl;
	}
	cout << endl;
}

void mznPrint (int N, int C, int H, int P, vvi &e, vvi &t, vvi &p)
{
	cout << "N = " << N << ";" << endl;
	cout << "C = " << C << ";" << endl;
	cout << "H = " << H << ";" << endl;
	cout << "P = " << P << ";" << endl;
	cout << endl;
	cout << "e = " ;
	mznvviPrint(e);
	cout << "t = " ;
	mznvviPrint(t);
	cout << "p = " ;
	for(auto& g : p)
	{
		g[1] += 1;
		g[2] += 1;
	}
	mznvviPrint(p);

	// exit(0);
	// vviPrint(t);
}

void glpvviPrint (vvi &Array)
{
	// int Ordem = Array.size();
	for (int k = 0; k < Array[0].size(); ++k)
		cout << setw(5) << k+1;
	cout << " :=" << endl;

	for (int i = 0; i < Array.size(); ++i)
	{
		// if ( i == 0 )
		// 	cout << "[|";
		// else
		cout << setw(10) << i+1 << " ";
		for (int j = 0; j < Array[i].size(); ++j)
		{
			cout.precision(2);
			cout << setw(4) << Array[i][j] << " ";
			// if ( j != Array[i].size()-1 )
			// 	cout << " ";
			// else
			// 	cout << " |";
		}
		if ( i != Array.size()-1 )
			cout << endl;
		else
			cout << ";" << endl;
	}
	cout << endl;
}

void glpPrint (int N, int C, int H, int P, vvi &e, vvi &t, vvi &p)
{
	cout << "data;" << endl << endl;
	cout << "param N := " << N << ";" << endl;
	cout << "param C := " << C << ";" << endl;
	cout << "param H := " << H << ";" << endl;
	cout << "param P := " << P << ";" << endl;
	cout << endl;
	cout << "param e : " ;
	glpvviPrint(e);
	cout << "param t : " ;
	glpvviPrint(t);
	cout << "param p : " ;
	for(auto& g : p)
	{
		g[1] += 1;
		g[2] += 1;
	}
	glpvviPrint(p);
	cout << "end;" << endl << endl;
}

template <typename T>
void printAuto(T vetor) {
	for (auto &i : vetor)
		// cout << i << " - ";
		cout << setw(2) << i << " - ";
	cout << endl;
}

/***************************************************************************
					Main function             
***************************************************************************/

int main (int argc, char* argv[])
{
	// user input
	// cities qty
	N = 10;
	// vehicle capacity except salesman
	int C = 3;
	//  symmetrical or asymmetrical (valid only for edges)
	// bool isSym = true;
	bool isSym = false;
	// output filename
	string out = "";

	int opt;
	while ( (opt = getopt(argc, argv, "n:c:s:o:")) != -1)
	{
		// SHOW(opt);
		switch (opt) {
			// SHOW(opt);
			case 'n':
				N = atof(optarg);
				break;
			case 'c':
				// SHOW(optarg);
				C = atof(optarg);
				break;
			case 's':
				// SHOW(optarg);
				isSym = atof(optarg);
				break;
			case 'o':
				// SHOW(optarg);
				out = optarg;
				break;
			default: /* '?' */
				fprintf(stderr, "Usage: %s [-n #CITIES] [-c VEHICLE CAPACITY] [-s SYMMETRICAL] [-o OUTPUT] \n", argv[0]);
			exit(EXIT_FAILURE);
		}
	}

	if (argc != 9)
	{
		fprintf(stderr, "Usage: %s [-n #CITIES] [-c VEHICLE CAPACITY] [-s SYMMETRICAL] [-o OUTPUT] \n", argv[0]);
		exit(EXIT_FAILURE);
	}

	// edge
	int eMin = 100;
	int eMax = 250;

	// toll percentage (applied to an edge)
	int tpMin = 25;
	int tpMax = 200;
	
	// tolls qty is a percentage of edges qty N*(N-1)/2
	// fixed
	// double S = 50; // [10,20,25]
	// S /= 100;
	// int H = S*(N*(N-1)/2);

	// max qty of tolls
	int E = N*(N-1)/2;
	// double S = mersenne(15,50)/100;
	// S /= 100;
	int Hmax = mersenne(15,50)*E/100;  // acop < 5%
	// int Hmax = mersenne(20,60)*E/100;  // acop < 5%
	if (Hmax < 3)
		Hmax = 3;
	int H = 0;


	// SHOW(S) SHOW(H)
	// now lets define Hmax elements to be non-zero, max
	// int Hmax = mersenne(15,50)*E/100;
	// int H = 0;

	// SHOW(Hmax) markhere();

	// min passenger rate percentage (applied onto MST cost)
	int mprpMin = 10;

	// max passenger rate percentage (applied onto MST cost)
	// predefined schemes in function of C
	int mprpMax;
	switch (C)
	{
		case 3: mprpMax = 20; break;
		case 4: mprpMax = 15; break;
	 	case 5: mprpMax = 15; break;
	 	case 6: mprpMax = 15; break;
	}

	// max qty of passengers available at a city
	// int maxQtyP = 4*C;
	// total qty of passengers
	int P = 0;


	// indexes
	vi count(N);
	int k = 0;
	for (auto &i : count)
	{
		i = k;
		k++;
	}

	vvi eMatrix, tMatrix, pMatrix;

	// eMatrix - matrix of edges' cost
	for (const auto &i : count)
	{
		vi line;
		for (const auto &j : count)
			if (i != j)
				line.push_back(mersenne(eMin,eMax));
			else
				line.push_back(0);
		eMatrix.push_back(line);
	}

	if (isSym)
		for (const auto &i : count)
			for (const auto &j : count)
				eMatrix[j][i] = eMatrix[i][j];

	graph = eMatrix;

	// tMatrix - matrix of tolls' cost
	// all zeroed
	for (const auto &i : count)
	{
		vi line(N,0);
		tMatrix.push_back(line);
	}

	/*
	// tolls be all united =/ 
	for (const auto &i : count)
		for (const auto &j : count)
			if ( i > j )
			{
				auto limite = (eMin+eMax)/2;
				auto e = eMatrix[i][j];
				auto dif = limite - e;
				auto difMax = (eMin+eMax)/2 - eMin;	// 75

				// < 175
				if ( i != j && dif > 0 )
				{
					int percentToll = 0;
					// 100-125
					if ( dif > (2/3) * difMax )	
						percentToll = mersenne(60,80);
					else
						// 125-150
						if ( dif > (1/3) * difMax ) // 125-150
							percentToll = mersenne(40,60);
						// 150-175
						else
							percentToll = mersenne(20,40);

					if ( mersenne(1,100) <= percentToll && H < Hmax )
					{
						// toll must be related to both
						auto eMij = eMatrix[i][j];
						auto eMji = eMatrix[j][i];
						auto eMed = ( eMij + eMji ) / 2;
						// fairness tests: toll >= 100 and edge+toll <= 2*eMax
						auto t = INF;
						while ( t < 100 || eMij + t > 2 * eMax || eMji + t > 2 * eMax)
						{
							double percent = mersenne(tpMin,tpMax);
							t = eMed * percent/100;
						}
						tMatrix[j][i] = tMatrix[i][j] = t;
						H++;
					}
				}
			}
	*/

	// code to prevent a TSP optimal solution from being a TSPMPHOP optimal solution 
	// lets make edges < (eMin+eMax)/2 = limit to have a corresponding non-zero toll cost with a chance
	// select all edges on the upper triangle with a big enough cost
	auto limit = (eMin+eMax)/2; // 175
	vpi selEdges;
	for (const auto &i : count)
		for (const auto &j : count)
			if ( i > j && eMatrix[i][j] < limit )
				selEdges.push_back({i,j});		

	// prevent all tolls to be united at small values of i and j
	random_shuffle( selEdges.begin(), selEdges.end() );

	// now we make tolls until Hmax is reached - tolls are always symmetrical
	auto difMax = limit - eMin;	// 75
	auto sumLimit = eMax * double(tpMax) / 100;// SHOW(sumLimit)
	for (int l = 0; l < selEdges.size(); ++l)
	{
		if ( H < Hmax )
		{
			auto i = selEdges[l].first;
			auto j = selEdges[l].second;

			// toll must be related to both in case its assymetrical
			auto eMij = eMatrix[i][j];
			auto eMji = eMatrix[j][i];
			auto dif = limit - min(eMij, eMji);
			int probToll = 0;
			if ( dif > (2/3) * difMax )	         // 100-125
				probToll = mersenne(70,90);
			else
				if ( dif > (1/3) * difMax )      // 125-150
					probToll = mersenne(50,70);
				else                             // 150-175
					probToll = mersenne(30,50);

			if ( mersenne(1,100) <= probToll )
			{
				// fairness tests: toll >= 100 and edge+toll <= 2*eMax
				auto t = INF;
				// auto eMed = ( eMij + eMji ) / 2;
				// while ( t < limit || eMij + t > sumLimit || eMji + t > sumLimit) // acop < 3%
				while ( t < limit || t > sumLimit )
					t = max(eMij, eMji) * double(mersenne(tpMin,tpMax)) / 100; // SHOW(t) markhere();
				tMatrix[j][i] = tMatrix[i][j] = t;
				H++;
			}
		}
		else
			l = selEdges.size();
	}



	// MST
	Graph PCV(N, E);
	for (const auto &i : count)
		for (const auto &j : count)
			PCV.addEdge(i,j,eMatrix[i][j]);
	// mst_wt = PCV.kruskalMST();
	// cout << "Weight of MST is " << mstCost << endl;
	auto mstCost = PCV.kruskalMST();
	// SHOW(mstCost)

 	// check condition

	for (const auto &orig : count)
	{
		// auto posO = findKey(orig, route);
		// auto posO = indexMap[ orig ];
		if (orig > 0)
		{
			// qty of passengers available at i - predefined schemes
			auto F = mersenne(1,3);
			int qtyP;
			switch (F)
			{
				// case 1: qtyP = mersenne(0*C,1*C); break;
				// case 2: qtyP = mersenne(1*C,2*C); break;
				// case 3: qtyP = mersenne(2*C,3*C); break;
		 		case 1: qtyP = mersenne(0*C,1*C); break;
		 		case 2: qtyP = mersenne(0*C,2*C); break;
		 	 	case 3: qtyP = mersenne(0*C,3*C); break;

	 	 		// case 1: qtyP = mersenne(0*C,3*C); break;
	 	 		// case 2: qtyP = mersenne(1*C,4*C); break;
	 	 		// case 3: qtyP = mersenne(3*C,5*C); break;
	 	 	 	

			}
			P += qtyP;

			int par = 0;

			int p = 0;
			while (p < qtyP)
			{
				int dest;
				dest = orig;
				while (dest == orig)
					dest = mersenne(0,N-1);

				// SHOW(orig) SHOW(dest) markhere();

				auto m = mersenne(mprpMin,mprpMax);
				// SHOW(m)
				auto maxRate = mstCost * m / 100;
				maxRate += mersenne(0,10); // prevent all zeros when mstCost ~ 1000

				p++;
				vi line;
				line.push_back(maxRate);
				line.push_back(orig);
				line.push_back(dest);
				pMatrix.push_back(line);
			}
		}
	}

	ofstream output;

	// cout << "OK" << endl;

	// string cl = "01";
	// string in = "01";

	// teste
	// cppPrint(N, C, H, P, eMatrix, tMatrix, pMatrix);
	// SHOW(E) SHOW(H) SHOW(Hmax) markhere();

	cout.rdbuf(output.rdbuf());	// don't print them out

	output.open( "in/txt/" + out + ".txt", ios::out | ios::app );
	cppPrint(N, C, H, P, eMatrix, tMatrix, pMatrix);
	output.close();

	output.open( "in/dzn/" + out + ".dzn", ios::out | ios::app );
	mznPrint(N, C, H, P, eMatrix, tMatrix, pMatrix);
	output.close();

	output.open( "in/dat/" + out + ".dat", ios::out | ios::app );
	glpPrint(N, C, H, P, eMatrix, tMatrix, pMatrix);
	output.close();

	exit(0);
}